def get():
    # todo, implement this to return the right connection for the child_class
    # connection = prom.connection.get(child_klass)
    pass

def set():
    # todo, set a connection for a child orm class using interface
#prom.connection.set(
#    orm=klass,
#    interface=Interface,
#    dbname="",
#    username="",
#    password=""
#    host="",
#    port=""
#)
    pass




"""
Handle standard query creation

example --

    query.table("table_name").is_foo(1).desc_bar().set_limit(10).set_page(2).get()

"""
#    query.table("table_name").is_foo(1).desc_bar().limit(10).page(2).get()
#    query.table("table_name").is_foo(1).desc_bar().set_limit(10).set_page(2).get()
#    query.table("table_name").is_foo(1).desc_bar().with_limit(10).with_page(2).get()
#    query.table("table_name").is_foo(1).desc_bar().use_limit(10).use_page(2).get()
#    query.table("table_name").is_foo(1).desc_bar().limit_to(10).on_page(2).with_offset(5).get()

class Query(object):

    def __init__(self, orm=None, *args, **kwargs):

        # needed to use the db querying methods like get(), if you just want to build
        # a query then you don't need to bother passing this in
        self.orm = orm

        self.fields = []
        self.fields_where = []
        self.fields_sort = []
        self.bounds = {}
        self.args = args
        self.kwargs = kwargs

    def set_field(self, field_name, field_val=None):
        """
        set a field into .fields attribute

        this has a dual role, in select queries, these are the select fields, but in insert/update
        queries, these are the fields that will be inserted/updated into the db
        """
        self.fields.append([field_name, field_val])
        return self

    def set_fields(self, fields=None, **fields_kwargs):
        """
        completely replaces the current .fields with fields and fields_kwargs combined
        """
        if not fields: fields = {}
        fields.update(fields_kwargs)
        self.fields = fields
        return self

    def is_field(self, field_name, field_val):
        self.fields_where.append(["is", field_name, field_val])
        return self

    def not_field(self, field_name, field_val):
        self.fields_where.append(["not", field_name, field_val])
        return self

    def between_field(self, field_name, low, high):
        self.lte_field(field_name, low)
        self.gte_field(field_name, high)
        return self

    def lte_field(self, field_name, field_val):
        self.fields_where.append(["lte", field_name, field_val])
        return self

    def lt_field(self, field_name, field_val):
        self.fields_where.append(["lt", field_name, field_val])
        return self

    def gte_field(self, field_name, field_val):
        self.fields_where.append(["gte", field_name, field_val])
        return self

    def gt_field(self, field_name, field_val):
        self.fields_where.append(["gt", field_name, field_val])
        return self

    def in_field(self, field_name, *field_vals):
        """
        NOTE -- if you have a list of values, it has to be passed in using the *, in_field("name", *list_of_values)
        """
        self.fields_where.append(["in", field_name, field_vals])
        return self

    def nin_field(self, field_name, *field_vals):
        """
        NOTE -- if you have a list of values, it has to be passed in using the *, nin_field("name", *list_of_values)
        """
        self.fields_where.append(["nin", field_name, field_vals])
        return self

    def sort_field(self, field_name, direction):
        if direction > 0:
            direction = 1
        elif direction < 0:
            direction = -1
        else:
            raise ValueError("direction {} is undefined".format(direction))

        self.fields_sort.append([direction, field_name])
        return self

    def asc_field(self, field_name):
        self.sort_field(field_name, 1)
        return self

    def desc_field(self, field_name):
        self.sort_field(field_name, -1)
        return self

    def __getattr__(self, method_name):

        command, field_name = self._split_method(method_name)

        def callback(*args, **kwargs):
            field_method_name = "{}_field".format(command)
            command_field_method = None

            if field_method_name in self.__class__.__dict__:
                command_field_method = getattr(self, field_method_name)
            else:
                raise AttributeError('No "{}" method derived from "{}"'.format(field_method_name, method_name))

            return command_field_method(field_name, *args, **kwargs)

        return callback

    def _split_method(self, method_name):
        command, field_name = method_name.split(u"_", 1)
        return command, field_name

    def set_limit(self, limit):
        self.bounds['limit'] = int(limit)
        return self

    def set_offset(self, offset):
        self.bounds.pop("page", None)
        self.bounds['offset'] = int(offset)
        return self

    def set_page(self, page):
        self.bounds.pop("offset", None)
        self.bounds['page'] = int(page)
        return self

    def get_bounds(self):

        limit = offset = page = limit_paginate = 0
        if "limit" in self.bounds and self.bounds["limit"] > 0:
            limit = self.bounds["limit"]
            limit_paginate = limit + 1

        if "offset" in self.bounds:
            offset = self.bounds["offset"]
            offset = offset if offset >= 0 else 0

        else:
            if "page" in self.bounds:
                page = self.bounds["page"]
                page = page if page >= 1 else 1
                offset = (page - 1) * limit

        return (limit, offset, limit_paginate)

    def has_bounds(self):
        return len(self.bounds) > 0

    def get(self, limit=None, page=None):
        if limit is not None:
            self.set_limit(limit)
        if page is not None:
            self.set_page(page)

        for d in self._query('get'):
            yield self._create_orm(d)

    def get_one(self):
        """get one row from the db"""
        o = None
        d = self._query('get_one')
        if d:
            o = self._create_orm(d)
        return o

    def get_pk(self, field_val):
        """convenience method for running is__id(_id).get_one() since this is so common"""
        field_name = self.orm.schema.pk
        return self.is_field(field_name, field_val).get_one()

    def count(self):
        """return the count of the criteria"""
        return self._query('count')

    def set(self):
        """persist the .fields using .fields_where (if available)"""
        return self._query('set')

    def delete(self):
        """remove fields matching the where criteria"""
        return self._query('delete')

    def get_query(self, query_str, *query_args, **query_options):
        """
        use the interface query method to pass in your own raw query
        """
        i = self.orm.interface
        return i.query(query_str, *query_args, **query_options)

    def _query(self, method_name):
        i = self.orm.interface
        s = self.orm.schema
        return getattr(i, method_name)(s, self)

    def _create_orm(self, d):
        o = self.orm(**d)
        o.reset_modified()
        return o




"""
this holds orm specific stuff
"""
import inspect
import sys

from .query import Query
from . import get_interface
from .utils import classproperty, cachedclassproperty

class Orm(object):
    """
    this is the parent class of any model Orm class you want to create that can access the db

    NOTE -- you must set the schema class as a class property (not an instance property)

    example -- create a user class

        import prom

        class User(prom.Orm):

            schema = prom.Schema(
                "user_table_name",
                username=(str, True),
                password=(str, True)
                email=(str, True)
                unique_user=('username') # set a unique index on user
                index_email=('email') # set a normal index on email
            )

        # create a user
        u = User(username='foo', password='awesome_and_secure_pw_hash', email='foo@bar.com')
        u.set()

        # query for our new user
        u = User.query.is_username('foo').get_one()
        print u.username # foo
    """

    connection_name = ""
    """the name of the connection to use to retrieve the interface"""

    schema = None
    """the Schema() instance that this class will derive all its db info from"""

    @cachedclassproperty
    def interface(cls):
        """
        return an Interface instance that can be used to access the db

        this will do the work required to find the interface instance, and then set this
        property to that interface instance, so all subsequent calls to this property
        will return the Interface() directly without running this method again

        return -- Interface() -- the interface instance this Orm will use
        """
        i = get_interface(cls.connection_name)
        return i

    @cachedclassproperty
    def query_class(cls):
        """
        return the Query instance this class will use create Query instances to actually
        query the db

        To be as magical and helpful as possible, this will start at the child class and look
        for a ChildClassNameQuery defined in the same module, it will then move down through
        each parent class looking for a matching <ParentClassName>Query class defined in the 
        same module as the parent class. If you want to short circuit the auto-finding, you 
        can just set the query_class in the child class

        example -- set the Query class manually

            class Foo(Orm):
                query_class = module.Query

        like .interface, this is cached

        return -- Query -- the query class, not an instance, but the actaul class object, Query if
            a custom one isn't found
        """
        query_class = Query
        parents = inspect.getmro(cls)
        for parent in parents:
            parent_module_name = parent.__module__
            parent_class_name = '{}Query'.format(parent.__name__)
            #pout.v(parent_module_name, parent_class_name)
            if parent_module_name in sys.modules:
                parent_class = getattr(sys.modules[parent_module_name], parent_class_name, None)
                if parent_class:
                    query_class = parent_class
                    break

        return query_class

    @classproperty
    def query(cls):
        """
        return a new Query instance ready to make a db call using the child class

        example -- fluid interface

            results = Orm.query.is_foo('value').desc_bar().get()

        return -- Query() -- every time this is called a new query instance is created using cls.query_class
        """
        query_class = cls.query_class
        return query_class(orm=cls)

    @property
    def pk(self):
        """wrapper method to return the primary key, None if the primary key is not set"""
        return getattr(self, self.schema.pk, None)

    def __init__(self, **fields):

        self.reset_modified()

        # go through and set all the values for 
        for field_name, field_val in fields.iteritems():
            setattr(self, field_name, field_val)

    def __setattr__(self, field_name, field_val):

#        if 'modified_fields' not in self.__dict__:
#            self.__dict__['modified_fields'] = set()

        s = self.schema
        if field_name in s.fields:
            self.modified_fields.add(field_name)

        self.__dict__[field_name] = field_val

    def set(self):

        ret = False
        d = {}

        # get all the modified fields
        for field_name in self.modified_fields:
            d[field_name] = getattr(self, field_name)

        if d:
            q = self.query
            # check if we should update, or insert
            _id_name = self.schema.pk
            _id = self.pk
            if _id:
                q.is_field(_id_name, _id)

            q.set_fields(d)
            d = q.set()

            for field_name, field_val in d.iteritems():
                d[field_name] = setattr(self, field_name, field_val)

            # orm values have now been persisted, so reset
            self.modified_fields = set()
            ret = True

        return ret

    def delete(self):

        ret = False
        q = self.query
        _id = self.pk
        _id_name = self.schema.pk
        if _id:
            self.query.is_field(_id_name, _id).delete()
            # get rid of _id
            delattr(self, _id_name)

            # mark all the fields that still exist as modified
            self.reset_modified()
            for field_name in self.schema.fields:
                if hasattr(self, field_name):
                    self.modified_fields.add(field_name)

            ret = True

        return ret

    def is_modified(self):
        """true if a field has been changed from its original value, false otherwise"""
        return len(self.modified_fields) > 0

    def reset_modified(self):
        """
        reset field modification tracking

        this is handy for when you are loading a new Orm with the results from a query and
        you don't want set() to do anything, you can Orm(**fields) and then orm.reset_modified() to
        clear all the passed in fields from the modified list
        """
#        if b:
#            if len(self.modified_fields) == 0:
#                for field_name in self.schema.fields:
#                    if not field_name.startwith('_'): # ignore the magic fields
#                        self.modified_fields.add(field_name)
#        else:
        self.modified_fields = set()

